@@ -54,6 +54,13 @@ SCM set_numlock_wrapper (SCM x);
 SCM set_scrolllock_wrapper (SCM x);
 SCM set_capslock_wrapper (SCM x);
 SCM xbindkey_wrapper(SCM key, SCM cmd);
+SCM xbindkey_function_wrapper(SCM key, SCM fun);
+SCM remove_xbindkey_wrapper(SCM key);
+SCM run_command_wrapper (SCM command);
+SCM grab_all_keys_wrapper (void);
+SCM ungrab_all_keys_wrapper (void);
+SCM remove_all_keys_wrapper (void);
+SCM debug_info_wrapper (void);
 #endif
 
 
@@ -64,9 +71,9 @@ get_options (int argc, char **argv)
   int i;
   char *home;
 
-  strcpy (rc_file, "");
+  strncpy (rc_file, "", sizeof(rc_file));
 #ifdef GUILE_FLAG
-  strcpy (rc_guile_file, "");
+  strncpy (rc_guile_file, "", sizeof (rc_guile_file));
 #endif
 
   verbose = 0;
@@ -159,7 +166,7 @@ get_options (int argc, char **argv)
       if (rc_file)
 	{
 	  strncpy (rc_file, home, sizeof (rc_file) - 20);
-	  strcat (rc_file, "/.xbindkeysrc");
+	  strncat (rc_file, "/.xbindkeysrc", sizeof (rc_file));
 	}
     }
 
@@ -171,7 +178,7 @@ get_options (int argc, char **argv)
       if (rc_guile_file)
 	{
 	  strncpy (rc_guile_file, home, sizeof (rc_guile_file) - 20);
-	  strcat (rc_guile_file, "/.xbindkeysrc.scm");
+	  strncat (rc_guile_file, "/.xbindkeysrc.scm", sizeof (rc_guile_file));
 	}
     }
 #endif
@@ -316,73 +323,134 @@ show_defaults_rc (void)
 static void
 show_defaults_guile_rc (void)
 {
-  printf (";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
-  printf ("; Start of xbindkeys configuration ;\n");
-  printf (";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
-  printf ("; This configuration is guile based.\n");
-  printf ("; any functions that work in guile will work here.\n");
-  printf ("; see EXTRA FUNCTIONS:\n");
+  printf (";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
+  printf (";; Start of xbindkeys configuration ;\n");
+  printf (";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
+  printf (";; This configuration is guile based.\n");
+  printf (";; any functions that work in guile will work here.\n");
+  printf (";; see EXTRA FUNCTIONS:\n");
   printf ("\n");
-  printf ("; Version: %s\n", PACKAGE_VERSION);
+  printf (";; Version: %s\n", PACKAGE_VERSION);
   printf ("\n");
-  printf ("; If you edit this file, do not forget to uncomment any lines\n");
-  printf ("; that you change.\n");
-  printf ("; The semicolon(;) symbol may be used anywhere for comments.\n");
+  printf (";; If you edit this file, do not forget to uncomment any lines\n");
+  printf (";; that you change.\n");
+  printf (";; The semicolon(;) symbol may be used anywhere for comments.\n");
   printf ("\n");
-  printf ("; To specify a key, you can use 'xbindkeys --key' or\n");
-  printf ("; 'xbindkeys --multikey' and put one of the two lines in this file.\n");
+  printf (";; To specify a key, you can use 'xbindkeys --key' or\n");
+  printf (";; 'xbindkeys --multikey' and put one of the two lines in this file.\n");
   printf ("\n");
-  printf ("; A list of keys is in /usr/include/X11/keysym.h and in\n");
-  printf ("; /usr/include/X11/keysymdef.h\n");
-  printf ("; The XK_ is not needed.\n");
+  printf (";; A list of keys is in /usr/include/X11/keysym.h and in\n");
+  printf (";; /usr/include/X11/keysymdef.h\n");
+  printf (";; The XK_ is not needed.\n");
   printf ("\n");
-  printf ("; List of modifier:\n");
-  printf (";   Release, Control, Shift, Mod1 (Alt), Mod2 (NumLock),\n");
-  printf (";   Mod3 (CapsLock), Mod4, Mod5 (Scroll).\n");
+  printf (";; List of modifier:\n");
+  printf (";;   Release, Control, Shift, Mod1 (Alt), Mod2 (NumLock),\n");
+  printf (";;   Mod3 (CapsLock), Mod4, Mod5 (Scroll).\n");
   printf ("\n");
   printf ("\n");
-  printf ("; The release modifier is not a standard X modifier, but you can\n");
-  printf ("; use it if you want to catch release instead of press events\n");
+  printf (";; The release modifier is not a standard X modifier, but you can\n");
+  printf (";; use it if you want to catch release instead of press events\n");
   printf ("\n");
-  printf ("; By defaults, xbindkeys does not pay attention to modifiers\n");
-  printf ("; NumLock, CapsLock and ScrollLock.\n");
-  printf ("; Uncomment the lines below if you want to use them.\n");
-  printf ("; To dissable them, call the functions with #f\n");
+  printf (";; By defaults, xbindkeys does not pay attention to modifiers\n");
+  printf (";; NumLock, CapsLock and ScrollLock.\n");
+  printf (";; Uncomment the lines below if you want to use them.\n");
+  printf (";; To dissable them, call the functions with #f\n");
   printf ("\n");
   printf ("\n");
-  printf (";EXTRA FUNCTIONS:\n");
-  printf (";(set-numlock! #t)\n");
-  printf (";(set-scrolllock! #t)\n");
-  printf (";(set-capslock! #t)\n");
-  printf (";(xbindkey key \"foo-bar-command [args]\")\n");
-  printf (";(xbindkey '(modifier* key) \"foo-bar-command [args]\")\n");
-  printf (";that is, xbindkey can take a list of modifiers ended by a key\n");
-  printf (";                  or it can just take a plain key.\n");
+  printf (";;EXTRA FUNCTIONS:\n");
+  printf (";;(set-numlock! #t)\n");
+  printf (";;(set-scrolllock! #t)\n");
+  printf (";;(set-capslock! #t)\n");
+  printf (";;that is, xbindkey can take a list of modifiers ended by a key\n");
+  printf (";;                  or it can just take a plain key.\n");
   printf ("\n");
-  printf ("; Examples of commands:\n");
+  printf (";;;;; Scheme API reference\n");
+  printf (";;;;\n");
+  printf (";; Optional modifier state:\n");
+  printf (";; (set-numlock! #f or #t)\n");
+  printf (";; (set-scrolllock! #f or #t)\n");
+  printf (";; (set-capslock! #f or #t)\n");
+  printf (";; \n");
+  printf (";; Shell command key:\n");
+  printf (";; (xbindkey key \"foo-bar-command [args]\")\n");
+  printf (";; (xbindkey '(modifier* key) \"foo-bar-command [args]\")\n");
+  printf (";; \n");
+  printf (";; Scheme function key:\n");
+  printf (";; (xbindkey-function key function-name-or-lambda-function)\n");
+  printf (";; (xbindkey-function '(modifier* key) function-name-or-lambda-function)\n");
+  printf (";; \n");
+  printf (";; Other functions:\n");
+  printf (";; (remove-xbindkey key)\n");
+  printf (";; (run-command \"foo-bar-command [args]\")\n");
+  printf (";; (grab-all-keys)\n");
+  printf (";; (ungrab-all-keys)\n");
+  printf (";; (remove-all-keys)\n");
+  printf (";; (debug)\n");
+  printf ("\n");
+  printf ("\n");
+  printf (";; Examples of commands:\n");
   printf ("\n");
   printf ("(xbindkey '(control shift q) \"xbindkeys_show\")\n");
   printf ("\n");
-  printf ("; set directly keycode (here control + f with my keyboard)\n");
+  printf (";; set directly keycode (here control + f with my keyboard)\n");
   printf ("(xbindkey '(\"m:0x4\" \"c:41\") \"xterm\")\n");
   printf ("\n");
-  printf ("; specify a mouse button\n");
+  printf (";; specify a mouse button\n");
   printf ("(xbindkey '(control \"b:2\") \"xterm\")\n");
   printf ("\n");
-  printf (";(xbindkey '(shift mod2 alt s) \"xterm -geom 50x20+20+20\")\n");
+  printf (";;(xbindkey '(shift mod2 alt s) \"xterm -geom 50x20+20+20\")\n");
   printf ("\n");
   printf (";; set directly keycode (control+alt+mod2 + f with my keyboard)\n");
   printf ("(xbindkey '(alt \"m:4\" mod2 \"c:0x29\") \"xterm\")\n");
   printf ("\n");
-  printf ("; Control+Shift+a  release event starts rxvt\n");
-  printf (";(xbindkey '(release control shift a) \"rxvt\")\n");
+  printf (";; Control+Shift+a  release event starts rxvt\n");
+  printf (";;(xbindkey '(release control shift a) \"rxvt\")\n");
   printf ("\n");
-  printf ("; Control + mouse button 2 release event starts rxvt\n");
-  printf (";(xbindkey '(releace control \"b:2\") \"rxvt\")\n");
+  printf (";; Control + mouse button 2 release event starts rxvt\n");
+  printf (";;(xbindkey '(releace control \"b:2\") \"rxvt\")\n");
   printf ("\n");
-  printf (";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
-  printf ("; End of xbindkeys configuration ;\n");
-  printf (";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
+  printf ("\n");
+  printf (";; Extra features\n");
+  printf ("(xbindkey-function '(control a)\n");
+  printf ("		   (lambda ()\n");
+  printf ("		     (display \"Hello from Scheme!\")\n");
+  printf ("		     (newline)))\n");
+  printf ("\n");
+  printf ("(xbindkey-function '(shift p)\n");
+  printf ("		   (lambda ()\n");
+  printf ("		     (run-command \"xterm\")))\n");
+  printf ("\n");
+  printf ("\n");
+  printf (";; Double click test\n");
+  printf ("(xbindkey-function '(control w)\n");
+  printf ("		   (let ((count 0))\n");
+  printf ("		     (lambda ()\n");
+  printf ("		       (set! count (+ count 1))\n");
+  printf ("		       (if (> count 1)\n");
+  printf ("			   (begin\n");
+  printf ("			    (set! count 0)\n");
+  printf ("			    (run-command \"xterm\"))))))\n");
+  printf ("\n");
+  printf (";; Time double click test:\n");
+  printf (";;  - short double click -> run an xterm\n");
+  printf (";;  - long  double click -> run an rxvt\n");
+  printf ("(xbindkey-function '(shift w)\n");
+  printf ("		   (let ((time (current-time))\n");
+  printf ("			 (count 0))\n");
+  printf ("		     (lambda ()\n");
+  printf ("		       (set! count (+ count 1))\n");
+  printf ("		       (if (> count 1)\n");
+  printf ("			   (begin\n");
+  printf ("			    (if (< (- (current-time) time) 1)\n");
+  printf ("				(run-command \"xterm\")\n");
+  printf ("				(run-command \"rxvt\"))\n");
+  printf ("			    (set! count 0)))\n");
+  printf ("		       (set! time (current-time)))))\n");
+  printf ("\n");
+  printf ("\n");
+  printf (";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
+  printf (";; End of xbindkeys configuration ;;\n");
+  printf (";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;\n");
 
   exit (1);
 }
@@ -668,7 +736,7 @@ get_rc_file (void)
 		    }
 
 		  if (add_key (type, event_type, keysym, keycode, 
-			       button, modifier, command) != 0)
+			       button, modifier, command, 0) != 0)
 		    break;
 		}
 	    }
@@ -711,10 +779,17 @@ init_xbk_guile_fns (void)
 {
   if (verbose)
     printf("initializing guile fns...\n");
-  scm_make_gsubr("set-numlock!", 1, 0, 0, set_numlock_wrapper);
-  scm_make_gsubr("set-scrolllock!", 1, 0, 0, set_scrolllock_wrapper);
-  scm_make_gsubr("set-capslock!", 1, 0, 0, set_capslock_wrapper);
-  scm_make_gsubr("xbindkey", 2, 0, 0, xbindkey_wrapper);
+  scm_c_define_gsubr("set-numlock!", 1, 0, 0, set_numlock_wrapper);
+  scm_c_define_gsubr("set-scrolllock!", 1, 0, 0, set_scrolllock_wrapper);
+  scm_c_define_gsubr("set-capslock!", 1, 0, 0, set_capslock_wrapper);
+  scm_c_define_gsubr("xbindkey", 2, 0, 0, xbindkey_wrapper);
+  scm_c_define_gsubr("xbindkey-function", 2, 0, 0, xbindkey_function_wrapper);
+  scm_c_define_gsubr("remove-xbindkey", 1, 0, 0, remove_xbindkey_wrapper);
+  scm_c_define_gsubr("run-command", 1, 0, 0, run_command_wrapper);
+  scm_c_define_gsubr("grab-all-keys", 0, 0, 0, grab_all_keys_wrapper);
+  scm_c_define_gsubr("ungrab-all-keys", 0, 0, 0, ungrab_all_keys_wrapper);
+  scm_c_define_gsubr("remove-all-keys", 0, 0, 0, remove_all_keys_wrapper);
+  scm_c_define_gsubr("debug", 0, 0, 0, debug_info_wrapper);
   return 0;
 }
 
@@ -757,24 +832,15 @@ MAKE_MASK_WRAPPER(set_numlock_wrapper, numlock_mask);
 MAKE_MASK_WRAPPER(set_scrolllock_wrapper, scrolllock_mask);
 MAKE_MASK_WRAPPER(set_capslock_wrapper, capslock_mask);
 
-SCM xbindkey_wrapper(SCM key, SCM cmd)
+
+
+SCM extract_key (SCM key, KeyType_t *type, EventType_t *event_type,
+		 KeySym *keysym, KeyCode *keycode,
+		 unsigned int *button, unsigned int *modifier)
 {
-  KeyType_t type = SYM;
-  EventType_t event_type = PRESS;
-  KeySym keysym = 0;
-  KeyCode keycode = 0;
-  unsigned int button = 0;
-  unsigned int modifier = 0;
-  char *cmdstr, *str;
+  char *str;
   int len;
   
-  //Guile strings are not \0 terminated. hence we must copy.
-  cmdstr = malloc(sizeof(char) * SCM_LENGTH(cmd) + 1);
-  strncpy(cmdstr, SCM_CHARS(cmd), SCM_LENGTH(cmd));
-  cmdstr[SCM_LENGTH(cmd)] = '\0';
-  if(verbose)
-    printf("xbindkey_wrapper debug: cmd=%s.\n", cmdstr);
-  
   while(SCM_CONSP(key)){ //Iterate through the list (If it is a list)
     if(!SCM_CONSP(SCM_CDR(key))){ //if this is the last item
       key = SCM_CAR(key);  //go to that
@@ -795,24 +861,24 @@ SCM xbindkey_wrapper(SCM key, SCM cmd)
     //copied directly with some substitutions. ie. line2 -> str
     //Do whatever needs to be done with modifiers.
     if (strncasecmp (str, "control", len) == 0)
-      modifier |= ControlMask;
+      *modifier |= ControlMask;
     else if (strncasecmp (str, "shift", len) == 0)
-      modifier |= ShiftMask;
+      *modifier |= ShiftMask;
     else if (strncasecmp (str, "mod1", len) == 0
              || strncasecmp (str, "alt", len) == 0)
-      modifier |= Mod1Mask;
+      *modifier |= Mod1Mask;
     else if (strncasecmp (str, "mod2", len) == 0)
-      modifier |= Mod2Mask;
+      *modifier |= Mod2Mask;
     else if (strncasecmp (str, "mod3", len) == 0)
-      modifier |= Mod3Mask;
+      *modifier |= Mod3Mask;
     else if (strncasecmp (str, "mod4", len) == 0)
-      modifier |= Mod4Mask;
+      *modifier |= Mod4Mask;
     else if (strncasecmp (str, "mod5", len) == 0)
-      modifier |= Mod5Mask;
+      *modifier |= Mod5Mask;
     else if (strncasecmp (str, "release", len) == 0)
-      event_type = RELEASE;
+      *event_type = RELEASE;
     else if(strlen (str) > 2 && str[0] == 'm' && str[1] == ':'){
-      modifier |= strtol (str+2, (char **) NULL, 0);
+      *modifier |= strtol (str+2, (char **) NULL, 0);
       //this break have nothing to do here!
       //break
     }else{
@@ -842,42 +908,173 @@ SCM xbindkey_wrapper(SCM key, SCM cmd)
     {
       switch (str[0])
         {
-	  case 'b':
-	      type = BUTTON;
-	      button = strtol (str+2, (char **) NULL, 0);
-	      break;
-	  case 'c':
-	      type = CODE;
-	      keycode = strtol (str+2, (char **) NULL, 0);
-	      break;
-	  case 'm': //is a modifier so it is in the other part.
-	      printf("bad modifyer: %s.", str);
-	      printf("m: modifiers need be applied to keys\n");
-  	      return SCM_BOOL_F;
-	  default:
-	      printf("bad modifyer: %c: shoud be b:, c: or m: .\n", str[0]);
-  	      return SCM_BOOL_F;
+	case 'b':
+	  *type = BUTTON;
+	  *button = strtol (str+2, (char **) NULL, 0);
+	  break;
+	case 'c':
+	  *type = CODE;
+	  *keycode = strtol (str+2, (char **) NULL, 0);
+	  break;
+	case 'm': //is a modifier so it is in the other part.
+	  printf("bad modifyer: %s.", str);
+	  printf("m: modifiers need be applied to keys\n");
+	  return SCM_BOOL_F;
+	default:
+	  printf("bad modifyer: %c: shoud be b:, c: or m: .\n", str[0]);
+	  return SCM_BOOL_F;
         }
     }
   else //regular key
     {
-      type = SYM;
-      keysym = XStringToKeysym (str);
-      if (keysym == 0){
+      *type = SYM;
+      *keysym = XStringToKeysym (str);
+      if (*keysym == 0){
         printf("No keysym for key: %s\n", str);
   	return SCM_BOOL_F;
       }
     }
+
+  free(str); //these were used by add key and copied.
+
+  return SCM_BOOL_T;
+}
+
+
+
+SCM xbindkey_wrapper(SCM key, SCM cmd)
+{
+  KeyType_t type = SYM;
+  EventType_t event_type = PRESS;
+  KeySym keysym = 0;
+  KeyCode keycode = 0;
+  unsigned int button = 0;
+  unsigned int modifier = 0;
+  char *cmdstr;
+  
+  //Guile strings are not \0 terminated. hence we must copy.
+  cmdstr = malloc(sizeof(char) * SCM_LENGTH(cmd) + 1);
+  strncpy(cmdstr, SCM_CHARS(cmd), SCM_LENGTH(cmd));
+  cmdstr[SCM_LENGTH(cmd)] = '\0';
+  if(verbose)
+    printf("xbindkey_wrapper debug: cmd=%s.\n", cmdstr);
+
+  if (extract_key (key, &type, &event_type, &keysym, &keycode,
+		   &button, &modifier) == SCM_BOOL_F)
+    {
+      return SCM_BOOL_F;
+    }
+  
   if (add_key (type, event_type, keysym, keycode, 
-       	button, modifier, cmdstr) != 0)
+       	button, modifier, cmdstr, 0) != 0)
     {
       printf("add_key didn't return 0!!!\n");
       return SCM_BOOL_F;
     }
   
-  free(str); //these were used by add key and copied.
   free(cmdstr); //we may get rid of them!
 
   return SCM_UNSPECIFIED;
 }
+
+
+SCM xbindkey_function_wrapper (SCM key, SCM fun)
+{
+  KeyType_t type = SYM;
+  EventType_t event_type = PRESS;
+  KeySym keysym = 0;
+  KeyCode keycode = 0;
+  unsigned int button = 0;
+  unsigned int modifier = 0;
+  
+  if (extract_key (key, &type, &event_type, &keysym, &keycode,
+		   &button, &modifier) == SCM_BOOL_F)
+    {
+      return SCM_BOOL_F;
+    }
+
+  if (add_key (type, event_type, keysym, keycode, 
+	       button, modifier, NULL, fun) != 0)
+    {
+      printf("add_key didn't return 0!!!\n");
+      return SCM_BOOL_F;
+    }
+
+  return SCM_UNSPECIFIED;
+}
+
+
+
+SCM remove_xbindkey_wrapper (SCM key)
+{
+  KeyType_t type = SYM;
+  EventType_t event_type = PRESS;
+  KeySym keysym = 0;
+  KeyCode keycode = 0;
+  unsigned int button = 0;
+  unsigned int modifier = 0;
+  
+  if (extract_key (key, &type, &event_type, &keysym, &keycode,
+		   &button, &modifier) == SCM_BOOL_F)
+    {
+      return SCM_BOOL_F;
+    }
+
+  if (remove_key (type, event_type, keysym, keycode, button, modifier) != 0)
+    {
+      printf("remove_key didn't return 0!!!\n");
+      return SCM_BOOL_F;
+    }
+ 
+  
+  return SCM_UNSPECIFIED;
+}
+
+
+SCM run_command_wrapper (SCM command)
+{
+  char *cmdstr;
+  
+  cmdstr = malloc(sizeof(char) * SCM_LENGTH(command) + 1);
+  strncpy(cmdstr, SCM_CHARS(command), SCM_LENGTH(command));
+  cmdstr[SCM_LENGTH(command)] = '\0';
+
+  run_command (cmdstr);
+
+  free(cmdstr);
+
+  return SCM_UNSPECIFIED;
+}
+
+SCM grab_all_keys_wrapper (void)
+{
+  grab_keys (current_display);
+  
+  return SCM_UNSPECIFIED;
+}
+
+
+SCM ungrab_all_keys_wrapper (void)
+{
+  ungrab_all_keys (current_display);
+  
+  return SCM_UNSPECIFIED;
+}
+
+SCM remove_all_keys_wrapper (void)
+{
+  close_keys ();
+  
+  return SCM_UNSPECIFIED;
+}
+
+
+SCM debug_info_wrapper (void)
+{
+  printf ("\nKeys = %p\n", keys);
+  printf ("nb_keys = %d\n", nb_keys);
+  
+  return SCM_UNSPECIFIED;
+}
+
 #endif
