@@ -42,12 +42,19 @@ void end_it_all (Display * d);
 static Display *start (char *);
 static void event_loop (Display *);
 static int *null_X_error (Display *, XErrorEvent *);
+static void reload_rc_file (void);
 static void catch_HUP_signal (int sig);
 static void catch_CHLD_signal (int sig);
 static void start_as_daemon (void);
 
-/* only for Catch_HUP_Signal */
-static Display *local_dpy;
+
+Display *current_display;  // The current display
+
+extern char rc_file[512];
+#ifdef GUILE_FLAG
+extern char rc_guile_file[512];
+#endif
+
 
 
 
@@ -81,7 +88,7 @@ inner_main (int argc, char **argv)
   show_options ();
 
   d = start (display_name);
-  local_dpy = d;
+  current_display = d;
 
   get_offending_modifiers (d);
 
@@ -181,13 +188,8 @@ and export or setenv it!\n");
 void
 end_it_all (Display * d)
 {
-  int screen;
+  ungrab_all_keys (d);
 
-  for (screen = 0; screen < ScreenCount (d); screen++)
-    {
-      XUngrabKey (d, AnyKey, AnyModifier, RootWindow (d, screen));
-      XUngrabButton (d, AnyButton, AnyModifier, RootWindow (d, screen));
-    }
   close_keys ();
   XCloseDisplay (d);
 }
@@ -197,12 +199,13 @@ end_it_all (Display * d)
 static void
 adjust_display (XAnyEvent * xany)
 {
-  char* envstr = malloc (strlen(DisplayString(xany->display)) + 8 + 1);
+  size_t envstr_size = strlen(DisplayString(xany->display)) + 8 + 1;
+  char* envstr = malloc (envstr_size);
   XWindowAttributes attr;
   char* ptr;
   char buf[16];
 
-  sprintf (envstr, "DISPLAY=%s", DisplayString(xany->display));
+  snprintf (envstr, envstr_size, "DISPLAY=%s", DisplayString(xany->display));
 
   XGetWindowAttributes (xany->display, xany->window,  &attr);
 
@@ -214,7 +217,7 @@ adjust_display (XAnyEvent * xany)
     *ptr = '\0';
 
   snprintf (buf, sizeof(buf), ".%i", XScreenNumberOfScreen(attr.screen));
-  strcat (envstr, buf);
+  strncat (envstr, buf, 16);
 
   putenv (envstr);
 }
@@ -226,13 +229,56 @@ event_loop (Display * d)
 {
   XEvent e;
   int i;
+  struct stat rc_file_info;
+  time_t rc_file_changed;
+#ifdef GUILE_FLAG
+  time_t rc_guile_file_changed;
+#endif
+
+
 
   XSetErrorHandler ((XErrorHandler) null_X_error);
 
+  stat(rc_file, &rc_file_info);
+  rc_file_changed = rc_file_info.st_mtime;
+#ifdef GUILE_FLAG
+  stat (rc_guile_file, &rc_file_info);
+  rc_guile_file_changed = rc_file_info.st_mtime;
+#endif
+
+  
   while (True)
     {
       XNextEvent (d, &e);
 
+      // if the rc file has been modified, reload it
+      stat (rc_file, &rc_file_info);
+      if (rc_file_info.st_mtime != rc_file_changed)
+	{
+	  reload_rc_file ();
+	  if (verbose)
+	    {
+	      printf ("The rc file has been modified, reload it\n");
+	    }
+	  rc_file_changed = rc_file_info.st_mtime;
+	}
+
+#ifdef GUILE_FLAG
+      // if the rc guile file has been modified, reload it
+      stat (rc_guile_file, &rc_file_info);
+      if (rc_file_info.st_mtime != rc_guile_file_changed)
+	{
+	  reload_rc_file ();
+	  if (verbose)
+	    {
+	      printf ("The rc file has been modified, reload it\n");
+	    }
+	  rc_guile_file_changed = rc_file_info.st_mtime;
+	}
+#endif
+
+      
+
       switch (e.type)
 	{
 	case KeyPress:
@@ -380,11 +426,11 @@ null_X_error (Display * d, XErrorEvent * e)
   already = 1;
 
   printf ("\n*** Warning ***\n");
-  printf ("Please, verifie that there is not another program which\n");
-  printf ("captures already one of the keys captured by xbindkeys.\n");
-  printf ("It seem that there is a conflict in keys and xbindkeys\n");
-  printf ("can't grab all the keys defined in its configuration file.\n");
-
+  printf ("Please verify that there is not another program running\n");
+  printf ("which captures one of the keys captured by xbindkeys.\n");
+  printf ("It seems that there is a conflict, and xbindkeys can't\n");
+  printf ("grab all the keys defined in its configuration file.\n");
+  
 /*   end_it_all (d); */
 
 /*   exit (-1); */
@@ -395,19 +441,19 @@ null_X_error (Display * d, XErrorEvent * e)
 
 
 static void
-catch_HUP_signal (int sig)
+reload_rc_file (void)
 {
   int min, max;
   int screen;
 
-  XDisplayKeycodes (local_dpy, &min, &max);
+  XDisplayKeycodes (current_display, &min, &max);
 
   if (verbose)
-    printf ("Catch HUP signal -> reread RC file\n");
+    printf ("Reload RC file\n");
 
-  for (screen = 0; screen < ScreenCount (local_dpy); screen++)
+  for (screen = 0; screen < ScreenCount (current_display); screen++)
     {
-      XUngrabKey (local_dpy, AnyKey, AnyModifier, RootWindow (local_dpy, screen));
+      XUngrabKey (current_display, AnyKey, AnyModifier, RootWindow (current_display, screen));
     }
   close_keys ();
 
@@ -417,14 +463,21 @@ catch_HUP_signal (int sig)
 #endif
       if (get_rc_file () != 0)
 	{
-	  end_it_all (local_dpy);
+	  end_it_all (current_display);
 	  exit (-1);
 	}
 #ifdef GUILE_FLAG
     }
 #endif
 
-  grab_keys (local_dpy);
+  grab_keys (current_display);
+}
+
+
+static void
+catch_HUP_signal (int sig)
+{
+  reload_rc_file ();
 }
 
 
